@@ -4,7 +4,7 @@
  * Copyright (C) 2017, 2018, 2019 Giuseppe Scrivano <giuseppe@scrivano.org>
  * crun is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
- * the Free Software Foundation; either version 3 of the License, or
+ * the Free Software Foundation; either version 2.1 of the License, or
  * (at your option) any later version.
  *
  * crun is distributed in the hope that it will be useful,
@@ -62,6 +62,8 @@ static struct symlink_s cgroup_symlinks[] = {
   { NULL, NULL }
 };
 
+#define SYSTEMD_PROPERTY_PREFIX "org.systemd.property."
+
 #ifndef CGROUP2_SUPER_MAGIC
 # define CGROUP2_SUPER_MAGIC 0x63677270
 #endif
@@ -126,6 +128,7 @@ enable_controllers (runtime_spec_schema_config_linux_resources *resources, const
 {
   cleanup_free char *controllers = NULL;
   cleanup_free char *tmp_path = NULL;
+  size_t controllers_len = 0;
   bool has_hugetlb = false;
   bool has_memory = false;
   bool has_cpuset = false;
@@ -149,16 +152,21 @@ enable_controllers (runtime_spec_schema_config_linux_resources *resources, const
       has_hugetlb = resources->hugepage_limits_len;
     }
 
-  xasprintf (&controllers, "%s %s %s %s %s %s",
-             has_cpu ? "+cpu" : "",
-             has_io ? "+io" : "",
-             has_memory ? "+memory" : "",
-             has_pids ? "+pids" : "",
-             has_cpuset ? "+cpuset" : "",
-             has_hugetlb ? "+hugetlb" : "");
+  controllers_len = xasprintf (&controllers, "%s %s %s %s %s %s",
+                               has_cpu ? "+cpu" : "",
+                               has_io ? "+io" : "",
+                               has_memory ? "+memory" : "",
+                               has_pids ? "+pids" : "",
+                               has_cpuset ? "+cpuset" : "",
+                               has_hugetlb ? "+hugetlb" : "");
 
   xasprintf (&tmp_path, "%s/", path);
 
+  /* Activate the needed controllers in the root cgroup, but ignore errors here.  */
+  ret = write_file ("/sys/fs/cgroup/cgroup.subtree_control", controllers, controllers_len, err);
+  if (UNLIKELY (ret < 0))
+    crun_error_release (err);
+
   for (it = strchr (tmp_path + 1, '/'); it;)
     {
       cleanup_free char *cgroup_path = NULL;
@@ -169,15 +177,14 @@ enable_controllers (runtime_spec_schema_config_linux_resources *resources, const
 
       xasprintf (&cgroup_path, "/sys/fs/cgroup%s", tmp_path);
       ret = mkdir (cgroup_path, 0755);
-      if (ret < 0 && errno != EEXIST) {
+      if (UNLIKELY (ret < 0 && errno != EEXIST))
         return crun_make_error (err, errno, "create `%s`", cgroup_path);
-      }
 
       if (next_slash)
         {
           xasprintf (&subtree_control, "%s/cgroup.subtree_control", cgroup_path);
-          ret = write_file (subtree_control, controllers, strlen (controllers), err);
-          if (ret < 0)
+          ret = write_file (subtree_control, controllers, controllers_len, err);
+          if (UNLIKELY (ret < 0))
             {
               int e = crun_error_get_errno (err);
               if (e == EPERM || e == EACCES || e == EBUSY)
@@ -689,8 +696,149 @@ systemd_job_removed (sd_bus_message *m, void *userdata, sd_bus_error *error arg_
   return 0;
 }
 
+/* Parse a gvariant string.  Support only a subset of types, just enough for systemd .  */
+static int
+append_systemd_annotation (sd_bus_message *m,
+                           const char *name,
+                           size_t name_len,
+                           const char *value,
+                           libcrun_error_t *err)
+{
+  cleanup_free char *tmp_name = NULL;
+  uint32_t factor = 1;
+  const char *it;
+  int sd_err;
+
+  while (*value == ' ')
+    value++;
+
+  it = value;
+
+  /* If the name has the form NameSec, convert it to NameUSec.  */
+  if (name_len > 4
+      && name[name_len - 4] != 'U'
+      && name[name_len - 3] == 'S'
+      && name[name_len - 2] == 'e'
+      && name[name_len - 1] == 'c')
+    {
+      factor = 1000000;
+
+      tmp_name = xmalloc (name_len + 2);
+      memcpy (tmp_name, name, name_len - 3);
+      memcpy (tmp_name + name_len - 3, "USec", 5);
+
+      name = tmp_name;
+    }
+
+
+  if ((strcmp (it, "true") == 0) || (strcmp (it, "false") == 0))
+    {
+      bool b = *it == 't';
+
+      sd_err = sd_bus_message_append (m, "(sv)", name, "b", b);
+      if (UNLIKELY (sd_err < 0))
+        return crun_make_error (err, -sd_err, "sd-bus message append `%s`", name);
+
+      return 0;
+    }
+  else if (*it == '\'')
+    {
+      cleanup_free char *v_start = NULL;
+      char *end;
+
+      it = v_start = xstrdup (value);
+
+      end = strchr (it + 1, '\'');
+      if (end == NULL)
+        return crun_make_error (err, 0, "invalid variant `%s`", value);
+      *end = '\0';
+
+      sd_err = sd_bus_message_append (m, "(sv)", name, "s", it + 1);
+      if (UNLIKELY (sd_err < 0))
+        return crun_make_error (err, -sd_err, "sd-bus message append `%s`", name);
+
+      return 0;
+    }
+  else if (has_prefix (it, "uint64 "))
+    {
+      char *endptr = NULL;
+      uint64_t v;
+
+      errno = 0;
+      v = strtoull (it + sizeof ("uint64"), &endptr, 10);
+      if (UNLIKELY (errno != 0 || *endptr))
+        return crun_make_error (err, errno, "invalid value for `%s`", name);
+
+      sd_err = sd_bus_message_append (m, "(sv)", name, "t", (uint64_t)(v * factor));
+      if (UNLIKELY (sd_err < 0))
+        return crun_make_error (err, -sd_err, "sd-bus message append `%s`", name);
+
+      return 0;
+    }
+  else if (has_prefix (it, "int64 "))
+    {
+      char *endptr = NULL;
+      int64_t v;
+
+      errno = 0;
+      v = strtoll (it + sizeof ("int64"), &endptr, 10);
+      if (UNLIKELY (errno != 0 || *endptr))
+        return crun_make_error (err, errno, "invalid value for `%s`", name);
+
+      sd_err = sd_bus_message_append (m, "(sv)", name, "x", (int64_t)(v * factor));
+      if (UNLIKELY (sd_err < 0))
+        return crun_make_error (err, -sd_err, "sd-bus message append `%s`", name);
+
+      return 0;
+    }
+  else if (has_prefix (it, "uint32 "))
+    {
+      char *endptr = NULL;
+      uint32_t v;
+
+      errno = 0;
+      v = strtoul (it + sizeof ("uint32"), &endptr, 10);
+      if (UNLIKELY (errno != 0 || *endptr))
+        return crun_make_error (err, errno, "invalid value for `%s`", name);
+
+      sd_err = sd_bus_message_append (m, "(sv)", name, "u", (uint32_t)(v * factor));
+      if (UNLIKELY (sd_err < 0))
+        return crun_make_error (err, -sd_err, "sd-bus message append `%s`", name);
+
+      return 0;
+    }
+  else if (has_prefix (it, "int32 ") || strchr (it, ' ') == NULL)
+    {
+      char *endptr = NULL;
+      int32_t v;
+
+      /* If no type is specified, try to parse it as int32.  */
+
+      errno = 0;
+      if (has_prefix (it, "int32 "))
+        v = strtol (it + sizeof ("int32"), &endptr, 10);
+      else
+        v = strtol (it, &endptr, 10);
+      if (UNLIKELY (errno != 0 || *endptr))
+        return crun_make_error (err, errno, "invalid value for `%s`", name);
+
+      sd_err = sd_bus_message_append (m, "(sv)", name, "i", (int32_t)(v * factor));
+      if (UNLIKELY (sd_err < 0))
+        return crun_make_error (err, -sd_err, "sd-bus message append `%s`", name);
+
+      return 0;
+    }
+
+  return crun_make_error (err, errno, "unknown type for `%s`", name);
+}
+
 static
-int enter_systemd_cgroup_scope (runtime_spec_schema_config_linux_resources *resources, const char *id, const char *slice, pid_t pid, libcrun_error_t *err)
+int enter_systemd_cgroup_scope (runtime_spec_schema_config_linux_resources *resources,
+                                json_map_string_string *annotations,
+                                const char *id,
+                                const char *slice,
+                                pid_t pid,
+                                libcrun_error_t *err)
 {
   sd_bus *bus = NULL;
   sd_bus_message *m = NULL;
@@ -801,11 +949,32 @@ int enter_systemd_cgroup_scope (runtime_spec_schema_config_linux_resources *reso
         }
     }
 
-  sd_err = sd_bus_message_append (m, "(sv)", "CollectMode", "s", "inactive-or-failed");
-  if (UNLIKELY (sd_err < 0))
+  if (annotations)
     {
-      ret = crun_make_error (err, -sd_err, "sd-bus message append CollectMode");
-      goto exit;
+      size_t prefix_len = sizeof (SYSTEMD_PROPERTY_PREFIX) - 1;
+      size_t i;
+
+      for (i = 0; i < annotations->len; i++)
+        {
+          size_t len;
+
+          if (! has_prefix (annotations->keys[i], SYSTEMD_PROPERTY_PREFIX))
+            continue;
+
+          len = strlen (annotations->keys[i]);
+          if (len < prefix_len + 3)
+            {
+              ret = crun_make_error (err, EINVAL, "invalid systemd property name `%s`",
+                                     annotations->keys[i]);
+              goto exit;
+            }
+
+          ret = append_systemd_annotation (m, annotations->keys[i] + prefix_len,
+                                           len - prefix_len,
+                                           annotations->values[i], err);
+          if (UNLIKELY (ret < 0))
+            goto exit;
+        }
     }
 
   sd_err = sd_bus_message_append (m, "(sv)", "Description", "s", "libcrun container");
@@ -997,7 +1166,10 @@ exit:
 #endif
 
 static int
-libcrun_cgroup_enter_internal (runtime_spec_schema_config_linux_resources *resources, int cgroup_mode, char **path, const char *cgroup_path, int manager, pid_t pid, const char *id, libcrun_error_t *err)
+libcrun_cgroup_enter_internal (runtime_spec_schema_config_linux_resources *resources,
+                               json_map_string_string *annotations arg_unused, int cgroup_mode,
+                               char **path, const char *cgroup_path, int manager, pid_t pid,
+                               const char *id, libcrun_error_t *err)
 {
   if (manager == CGROUP_MANAGER_DISABLED)
     {
@@ -1010,7 +1182,7 @@ libcrun_cgroup_enter_internal (runtime_spec_schema_config_linux_resources *resou
     {
       int ret;
 
-      ret = enter_systemd_cgroup_scope (resources, id, cgroup_path, pid, err);
+      ret = enter_systemd_cgroup_scope (resources, annotations, id, cgroup_path, pid, err);
       if (UNLIKELY (ret < 0))
         return ret;
 
@@ -1041,7 +1213,17 @@ libcrun_cgroup_enter_internal (runtime_spec_schema_config_linux_resources *resou
 }
 
 int
-libcrun_cgroup_enter (runtime_spec_schema_config_linux_resources *resources, int cgroup_mode, char **path, const char *cgroup_path, int manager, pid_t pid, uid_t root_uid, gid_t root_gid, const char *id, libcrun_error_t *err)
+libcrun_cgroup_enter (runtime_spec_schema_config_linux_resources *resources,
+                      json_map_string_string *annotations,
+                      int cgroup_mode,
+                      char **path,
+                      const char *cgroup_path,
+                      int manager,
+                      pid_t pid,
+                      uid_t root_uid,
+                      gid_t root_gid,
+                      const char *id,
+                      libcrun_error_t *err)
 {
   libcrun_error_t tmp_err = NULL;
   int rootless;
@@ -1061,7 +1243,7 @@ libcrun_cgroup_enter (runtime_spec_schema_config_linux_resources *resources, int
         return crun_make_error (err, errno, "cgroups in hybrid mode not supported, drop all controllers from cgroupv2");
     }
 
-  ret = libcrun_cgroup_enter_internal (resources, cgroup_mode, path, cgroup_path, manager, pid, id, err);
+  ret = libcrun_cgroup_enter_internal (resources, annotations, cgroup_mode, path, cgroup_path, manager, pid, id, err);
   if (LIKELY (ret == 0))
     {
       if (cgroup_mode == CGROUP_MODE_UNIFIED && (root_uid != (uid_t) -1 || root_gid != (gid_t) -1))
@@ -1090,6 +1272,36 @@ libcrun_cgroup_enter (runtime_spec_schema_config_linux_resources *resources, int
   return ret;
 }
 
+int
+libcrun_cgroup_is_container_paused (const char *cgroup_path, int cgroup_mode, bool *paused, libcrun_error_t *err)
+{
+  cleanup_free char *content = NULL;
+  cleanup_free char *path = NULL;
+  const char *state;
+  int ret;
+
+  if (cgroup_path == NULL || cgroup_path[0] == '\0')
+    return 0;
+
+  if (cgroup_mode == CGROUP_MODE_UNIFIED)
+    {
+      state = "1";
+      xasprintf (&path, "/sys/fs/cgroup/%s/cgroup.freeze", cgroup_path);
+    }
+  else
+    {
+      state = "FROZEN";
+      xasprintf (&path, "/sys/fs/cgroup/freezer/%s/freezer.state", cgroup_path);
+    }
+
+  ret = read_all_file (path, &content, NULL, err);
+  if (UNLIKELY (ret < 0))
+    return ret;
+
+  *paused = strstr (content, state) != NULL;
+  return 0;
+}
+
 static int
 libcrun_cgroup_pause_unpause_with_mode (const char *cgroup_path, int cgroup_mode, const bool pause, libcrun_error_t *err)
 {
@@ -1222,7 +1434,7 @@ libcrun_cgroup_read_pids (const char *path, bool recurse, pid_t **pids, libcrun_
 
     case CGROUP_MODE_HYBRID:
     case CGROUP_MODE_LEGACY:
-      xasprintf (&cgroup_path, "/sys/fs/cgroup/pids/%s", path);
+      xasprintf (&cgroup_path, "/sys/fs/cgroup/memory/%s", path);
       break;
 
     default:
@@ -1288,7 +1500,6 @@ libcrun_cgroup_destroy (const char *id, char *path, int systemd_cgroup, libcrun_
 {
   int ret;
   size_t i;
-  ssize_t path_len;
   int mode;
   const cgroups_subsystem_t *subsystems;
 
@@ -1319,59 +1530,36 @@ libcrun_cgroup_destroy (const char *id, char *path, int systemd_cgroup, libcrun_
   if (UNLIKELY (ret < 0))
     crun_error_release (err);
 
-  path_len = strlen (path);
-  while (1)
+  if (mode == CGROUP_MODE_UNIFIED)
     {
-      if (mode == CGROUP_MODE_UNIFIED)
+      cleanup_free char *cgroup_path = NULL;
+
+      xasprintf (&cgroup_path, "/sys/fs/cgroup/%s", path);
+      rmdir (cgroup_path);
+    }
+  else
+    {
+      for (i = 0; subsystems[i]; i++)
         {
           cleanup_free char *cgroup_path = NULL;
 
-          xasprintf (&cgroup_path, "/sys/fs/cgroup/%s", path);
-          if (rmdir (cgroup_path) < 0)
-            break;
+          if (mode == CGROUP_MODE_LEGACY && strcmp (subsystems[i], "unified") == 0)
+            continue;
+
+          xasprintf (&cgroup_path, "/sys/fs/cgroup/%s/%s", subsystems[i], path);
+
+          rmdir (cgroup_path);
         }
-      else
-        {
-          bool cleaned_any = false;
-
-          for (i = 0; subsystems[i]; i++)
-            {
-              cleanup_free char *cgroup_path = NULL;
-
-              if (mode == CGROUP_MODE_LEGACY && strcmp (subsystems[i], "unified") == 0)
-                continue;
-
-              xasprintf (&cgroup_path, "/sys/fs/cgroup/%s/%s", subsystems[i], path);
-
-              if (rmdir (cgroup_path) == 0)
-                cleaned_any = true;
-            }
-
-          if (!cleaned_any)
-            break;
-        }
-
-      if (path_len <= 1)
-       break;
-
-      for (; path_len > 1 && path[path_len] != '/'; path_len--);
-      if (path_len > 1)
-       path[path_len] = '\0';
-   }
+    }
 
   return 0;
 }
 
 /* The parser generates different structs but they are really all the same.  */
-struct throttling_s
-{
-    int64_t major;
-    int64_t minor;
-    uint64_t rate;
-};
+typedef runtime_spec_schema_defs_linux_block_io_device_throttle throttling_s;
 
 static int
-write_blkio_v1_resources_throttling (int dirfd, const char *name, struct throttling_s **throttling, size_t throttling_len, libcrun_error_t *err)
+write_blkio_v1_resources_throttling (int dirfd, const char *name, throttling_s **throttling, size_t throttling_len, libcrun_error_t *err)
 {
   char fmt_buf[128];
   size_t i;
@@ -1401,7 +1589,7 @@ write_blkio_v1_resources_throttling (int dirfd, const char *name, struct throttl
 }
 
 static int
-write_blkio_v2_resources_throttling (int fd, const char *name, struct throttling_s **throttling, size_t throttling_len, libcrun_error_t *err)
+write_blkio_v2_resources_throttling (int fd, const char *name, throttling_s **throttling, size_t throttling_len, libcrun_error_t *err)
 {
   char fmt_buf[128];
   size_t i;
@@ -1524,28 +1712,28 @@ write_blkio_resources (int dirfd, bool cgroup2, runtime_spec_schema_config_linux
         return crun_make_error (err, errno, "open io.max");
 
       ret = write_blkio_v2_resources_throttling (wfd, "rbps",
-                                                 (struct throttling_s **) blkio->throttle_read_bps_device,
+                                                 (throttling_s **) blkio->throttle_read_bps_device,
                                                  blkio->throttle_read_bps_device_len,
                                                  err);
       if (UNLIKELY (ret < 0))
         return ret;
 
       ret = write_blkio_v2_resources_throttling (wfd, "wbps",
-                                                 (struct throttling_s **) blkio->throttle_write_bps_device,
+                                                 (throttling_s **) blkio->throttle_write_bps_device,
                                                  blkio->throttle_write_bps_device_len,
                                                  err);
       if (UNLIKELY (ret < 0))
         return ret;
 
       ret = write_blkio_v2_resources_throttling (wfd, "riops",
-                                                 (struct throttling_s **) blkio->throttle_read_iops_device,
+                                                 (throttling_s **) blkio->throttle_read_iops_device,
                                                  blkio->throttle_read_iops_device_len,
                                                  err);
       if (UNLIKELY (ret < 0))
         return ret;
 
       ret = write_blkio_v2_resources_throttling (wfd, "wiops",
-                                                 (struct throttling_s **) blkio->throttle_write_iops_device,
+                                                 (throttling_s **) blkio->throttle_write_iops_device,
                                                  blkio->throttle_write_iops_device_len,
                                                  err);
       if (UNLIKELY (ret < 0))
@@ -1554,28 +1742,28 @@ write_blkio_resources (int dirfd, bool cgroup2, runtime_spec_schema_config_linux
   else
     {
       ret = write_blkio_v1_resources_throttling (dirfd, "blkio.throttle.read_bps_device",
-                                                 (struct throttling_s **) blkio->throttle_read_bps_device,
+                                                 (throttling_s **) blkio->throttle_read_bps_device,
                                                  blkio->throttle_read_bps_device_len,
                                                  err);
       if (UNLIKELY (ret < 0))
         return ret;
 
       ret = write_blkio_v1_resources_throttling (dirfd, "blkio.throttle.write_bps_device",
-                                                 (struct throttling_s **) blkio->throttle_write_bps_device,
+                                                 (throttling_s **) blkio->throttle_write_bps_device,
                                                  blkio->throttle_write_bps_device_len,
                                                  err);
       if (UNLIKELY (ret < 0))
         return ret;
 
       ret = write_blkio_v1_resources_throttling (dirfd, "blkio.throttle.read_iops_device",
-                                                 (struct throttling_s **) blkio->throttle_read_iops_device,
+                                                 (throttling_s **) blkio->throttle_read_iops_device,
                                                  blkio->throttle_read_iops_device_len,
                                                  err);
       if (UNLIKELY (ret < 0))
         return ret;
 
       ret = write_blkio_v1_resources_throttling (dirfd, "blkio.throttle.write_iops_device",
-                                                 (struct throttling_s **) blkio->throttle_write_iops_device,
+                                                 (throttling_s **) blkio->throttle_write_iops_device,
                                                  blkio->throttle_write_iops_device_len,
                                                  err);
       if (UNLIKELY (ret < 0))
