@@ -4,7 +4,7 @@
  * Copyright (C) 2017, 2018, 2019 Giuseppe Scrivano <giuseppe@scrivano.org>
  * crun is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
- * the Free Software Foundation; either version 3 of the License, or
+ * the Free Software Foundation; either version 2.1 of the License, or
  * (at your option) any later version.
  *
  * crun is distributed in the hope that it will be useful,
@@ -41,7 +41,6 @@
 #include <sys/epoll.h>
 #include <sys/socket.h>
 #include <grp.h>
-#include "sig2str.h"
 
 #ifdef HAVE_SYSTEMD
 # include <systemd/sd-daemon.h>
@@ -450,7 +449,7 @@ int unblock_signals (libcrun_error_t *err)
 
   memset (&act, 0, sizeof (act));
   act.sa_handler = SIG_DFL;
-  for (i = 0; i < SIGNUM_BOUND; i++)
+  for (i = 0; i < NSIG; i++)
     {
       ret = sigaction (i, &act, NULL);
       if (ret < 0 && errno != EINVAL)
@@ -641,6 +640,10 @@ container_init_setup (void *args, const char *notify_socket,
   if (UNLIKELY (ret < 0))
     return ret;
 
+  ret = libcrun_reopen_dev_null (err);
+  if (UNLIKELY (ret < 0))
+    return ret;
+
   if (clearenv ())
     return crun_make_error (err, errno, "clearenv");
 
@@ -676,14 +679,14 @@ container_init_setup (void *args, const char *notify_socket,
         }
     }
 
+  ret = setsid ();
+  if (UNLIKELY (ret < 0))
+    return crun_make_error (err, errno, "setsid");
+
   if (has_terminal)
     {
       cleanup_close int terminal_fd = -1;
 
-      ret = setsid ();
-      if (UNLIKELY (ret < 0))
-        return crun_make_error (err, errno, "setsid");
-
       fflush (stderr);
 
       terminal_fd = libcrun_set_terminal (container, err);
@@ -719,7 +722,7 @@ container_init_setup (void *args, const char *notify_socket,
     }
 
   if (def->process->user)
-    umask (def->process->user->umask);
+    umask (def->process->user->umask_present ? def->process->user->umask : 0022);
 
   if (def->process && !def->process->no_new_privileges)
     {
@@ -1424,7 +1427,7 @@ libcrun_container_run_internal (libcrun_container_t *container, libcrun_context_
 
   container->context = context;
 
-  if (!detach)
+  if (!detach || context->notify_socket)
     {
       ret = prctl (PR_SET_CHILD_SUBREAPER, 1, 0, 0, 0);
       if (UNLIKELY (ret < 0))
@@ -1498,7 +1501,8 @@ libcrun_container_run_internal (libcrun_container_t *container, libcrun_context_
   /* If we are root (either on the host or in a namespace), then chown the cgroup to root in the container user namespace.  */
   get_root_in_the_userns_for_cgroups (def, container->host_uid, container->host_gid, &root_uid, &root_gid);
 
-  ret = libcrun_cgroup_enter (def->linux ? def->linux->resources : NULL, cgroup_mode,
+  ret = libcrun_cgroup_enter (def->linux ? def->linux->resources : NULL, def->annotations,
+                              cgroup_mode,
                               &cgroup_path, def->linux ? def->linux->cgroups_path : "",
                               cgroup_manager, pid, root_uid, root_gid, context->id, err);
   if (UNLIKELY (ret < 0))
@@ -1947,6 +1951,7 @@ int
 libcrun_get_container_state_string (const char *id, libcrun_container_status_t *status, const char *state_root, const char **container_status, int *running, libcrun_error_t *err)
 {
   int ret, has_fifo = 0;
+  bool paused = false;
 
   ret = libcrun_is_container_running (status, err);
   if (UNLIKELY (ret < 0))
@@ -1961,10 +1966,25 @@ libcrun_get_container_state_string (const char *id, libcrun_container_status_t *
       has_fifo = ret;
     }
 
+  if (*running && !has_fifo)
+    {
+      int cgroup_mode;
+
+      cgroup_mode = libcrun_get_cgroup_mode (err);
+      if (cgroup_mode < 0)
+        return cgroup_mode;
+
+      ret = libcrun_cgroup_is_container_paused (status->cgroup_path, cgroup_mode, &paused, err);
+      if (UNLIKELY (ret < 0))
+        return ret;
+    }
+
   if (! *running)
     *container_status = "stopped";
   else if (has_fifo)
     *container_status = "created";
+  else if (paused)
+    *container_status = "paused";
   else
     *container_status = "running";
 
@@ -2165,9 +2185,27 @@ libcrun_container_exec (libcrun_context_t *context, const char *id, runtime_spec
       if (UNLIKELY (ret < 0))
         return ret;
 
-      for (i = 0; i < process->env_len; i++)
-        if (putenv (process->env[i]) < 0)
-          libcrun_fail_with_error ( errno, "putenv `%s`", process->env[i]);
+      ret = clearenv ();
+      if (UNLIKELY (ret < 0))
+        return crun_make_error (err, 0, "clearenv");
+
+      if (process->env_len)
+        {
+          for (i = 0; i < process->env_len; i++)
+            if (putenv (process->env[i]) < 0)
+              libcrun_fail_with_error ( errno, "putenv `%s`", process->env[i]);
+        }
+      else if (container->container_def->process->env_len)
+        {
+          char *e;
+
+          for (i = 0; i < container->container_def->process->env_len; i++)
+            {
+              e = container->container_def->process->env[i];
+              if (putenv (e) < 0)
+                libcrun_fail_with_error ( errno, "putenv `%s`", e);
+            }
+        }
 
       if (getenv ("HOME") == NULL)
         {
@@ -2259,7 +2297,7 @@ libcrun_container_exec (libcrun_context_t *context, const char *id, runtime_spec
         }
 
       if (process->user)
-        umask (process->user->umask);
+        umask (process->user->umask_present ? process->user->umask : 0022);
 
       TEMP_FAILURE_RETRY (write (pipefd1, "0", 1));
       TEMP_FAILURE_RETRY (close (pipefd1));
@@ -2426,3 +2464,127 @@ libcrun_container_unpause (libcrun_context_t *context, const char *id, libcrun_e
 
   return libcrun_container_unpause_linux (&status, err);
 }
+
+int
+libcrun_container_checkpoint (libcrun_context_t *context, const char *id,
+                              libcrun_checkpoint_restore_t *cr_options,
+                              libcrun_error_t *err)
+{
+  int ret;
+  const char *state_root = context->state_root;
+  libcrun_container_status_t status;
+  cleanup_free libcrun_container_t *container = NULL;
+
+  memset (&status, 0, sizeof (status));
+  ret = libcrun_read_container_status (&status, state_root, id, err);
+  if (UNLIKELY (ret < 0))
+    return ret;
+
+  ret = libcrun_is_container_running (&status, err);
+  if (UNLIKELY (ret < 0))
+    return ret;
+  if (ret == 0)
+    return crun_make_error (err, errno, "the container `%s` is not running",
+                            id);
+
+  ret = read_container_config_from_state (&container, state_root, id, err);
+  if (UNLIKELY (ret < 0))
+    return ret;
+  return libcrun_container_checkpoint_linux (&status, container, cr_options,
+                                             err);
+}
+
+int
+libcrun_container_restore (libcrun_context_t *context, const char *id,
+                           libcrun_checkpoint_restore_t *cr_options,
+                           libcrun_error_t *err)
+{
+  cleanup_free libcrun_container_t *container = NULL;
+  runtime_spec_schema_config_schema *def = NULL;
+  const char *state_root = context->state_root;
+  cleanup_free char *cgroup_path = NULL;
+  libcrun_container_status_t status;
+  int cgroup_mode, cgroup_manager;
+  uid_t root_uid = -1;
+  gid_t root_gid = -1;
+  int ret;
+
+  memset (&status, 0, sizeof (status));
+  ret = libcrun_read_container_status (&status, state_root, id, err);
+  if (UNLIKELY (ret < 0))
+    return ret;
+
+  ret = libcrun_is_container_running (&status, err);
+  if (UNLIKELY (ret < 0))
+    return ret;
+  if (ret == 1)
+    return crun_make_error (err, 0,
+                            "the container `%s` is not in 'stopped' state",
+                            id);
+
+  ret = read_container_config_from_state (&container, state_root, id, err);
+  if (UNLIKELY (ret < 0))
+    return ret;
+
+  ret = libcrun_container_restore_linux (&status, container, cr_options, err);
+  if (UNLIKELY (ret < 0))
+    return ret;
+
+  /* Now that the process has been restored, moved it into is cgroup again.
+   * The whole cgroup code is copied from libcrun_container_run_internal(). */
+  def = container->container_def;
+
+  cgroup_mode = libcrun_get_cgroup_mode (err);
+  if (cgroup_mode < 0)
+    return cgroup_mode;
+
+  cgroup_manager = CGROUP_MANAGER_CGROUPFS;
+  if (context->systemd_cgroup)
+    cgroup_manager = CGROUP_MANAGER_SYSTEMD;
+  else if (context->force_no_cgroup)
+    cgroup_manager = CGROUP_MANAGER_DISABLED;
+
+  /* If we are root (either on the host or in a namespace),
+   * then chown the cgroup to root in the container user namespace. */
+  get_root_in_the_userns_for_cgroups (def, container->host_uid,
+                                      container->host_gid, &root_uid,
+                                      &root_gid);
+
+  ret = libcrun_cgroup_enter (def->linux ? def->linux->resources : NULL,
+                              def->annotations,
+                              cgroup_mode, &cgroup_path,
+                              def->linux ? def->linux->cgroups_path : "",
+                              cgroup_manager, status.pid, root_uid, root_gid,
+                              context->id, err);
+  if (UNLIKELY (ret < 0))
+    return ret;
+
+  if (def->linux && def->linux->resources)
+    {
+      ret =
+        libcrun_update_cgroup_resources (cgroup_mode, def->linux->resources,
+                                         cgroup_path, err);
+      if (UNLIKELY (ret < 0))
+        return ret;
+    }
+
+  ret = write_container_status (container, context, status.pid,
+                                status.cgroup_path, status.created, err);
+  if (UNLIKELY (ret < 0))
+    return ret;
+
+  if (!cr_options->detach)
+    {
+      int wait_status;
+      ret = waitpid (status.pid, &wait_status, 0);
+      if (UNLIKELY (ret < 0))
+        return crun_make_error (err, errno,
+                                "waitpid failed for container '%s' with %d",
+                                id, ret);
+
+      if (WEXITSTATUS (wait_status))
+        return WEXITSTATUS (wait_status);
+    }
+
+  return 0;
+}
