@@ -4,7 +4,7 @@
  * Copyright (C) 2017, 2018, 2019 Giuseppe Scrivano <giuseppe@scrivano.org>
  * crun is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
- * the Free Software Foundation; either version 3 of the License, or
+ * the Free Software Foundation; either version 2.1 of the License, or
  * (at your option) any later version.
  *
  * crun is distributed in the hope that it will be useful,
@@ -72,6 +72,18 @@ typedef struct libcrun_container_status_s libcrun_container_status_t;
 typedef struct libcrun_container_s libcrun_container_t;
 typedef struct libcrun_context_s libcrun_context_t;
 
+struct libcrun_checkpoint_restore_s
+{
+  char *image_path;
+  char *work_path;
+  bool leave_running;
+  bool tcp_established;
+  bool shell_job;
+  bool ext_unix_sk;
+  bool detach;
+};
+typedef struct libcrun_checkpoint_restore_s libcrun_checkpoint_restore_t;
+
 libcrun_container_t *libcrun_container_load_from_file (const char *path, libcrun_error_t *err);
 
 libcrun_container_t *libcrun_container_load_from_memory (const char *json, libcrun_error_t *err);
@@ -104,4 +116,8 @@ int libcrun_container_pause (libcrun_context_t *context, const char *id, libcrun
 
 int libcrun_container_unpause (libcrun_context_t *context, const char *id, libcrun_error_t *err);
 
+int libcrun_container_checkpoint (libcrun_context_t *context, const char *id, libcrun_checkpoint_restore_t * cr_options, libcrun_error_t *err);
+
+int libcrun_container_restore (libcrun_context_t *context, const char *id, libcrun_checkpoint_restore_t * cr_options, libcrun_error_t *err);
+
 #endif
