@@ -4,7 +4,7 @@
  * Copyright (C) 2017, 2018, 2019 Giuseppe Scrivano <giuseppe@scrivano.org>
  * crun is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
- * the Free Software Foundation; either version 3 of the License, or
+ * the Free Software Foundation; either version 2.1 of the License, or
  * (at your option) any later version.
  *
  * crun is distributed in the hope that it will be useful,
@@ -39,6 +39,7 @@
 #include "terminal.h"
 #include "cgroup.h"
 #include "status.h"
+#include "criu.h"
 #include <sys/socket.h>
 #include <libgen.h>
 #include <sys/wait.h>
@@ -115,6 +116,9 @@ static struct linux_namespace_s namespaces[] =
     {"user", CLONE_NEWUSER},
 #ifdef CLONE_NEWCGROUP
     {"cgroup", CLONE_NEWCGROUP},
+#endif
+#ifdef CLONE_NEWTIME
+    {"time", CLONE_NEWTIME},
 #endif
     {NULL, 0}
   };
@@ -633,6 +637,11 @@ do_mount_cgroup_v1 (libcrun_container_t *container,
   char *from;
   cleanup_close int tmpfsdirfd = -1;
   char *saveptr = NULL;
+  bool has_cgroupns = false;
+
+#ifdef CLONE_NEWCGROUP
+  has_cgroupns = get_private_data (container)->unshare_flags & CLONE_NEWCGROUP;
+#endif
 
   subsystems = libcrun_get_cgroups_subsystems (err);
   if (UNLIKELY (subsystems == NULL))
@@ -659,7 +668,7 @@ do_mount_cgroup_v1 (libcrun_container_t *container,
       cleanup_free char *source_path = NULL;
       cleanup_free char *source_subsystem = NULL;
       cleanup_free char *subsystem_path = NULL;
-      char *subpath, *subsystem, *it;
+      char *subpath, *subsystem, *subsystem_fqn, *it;
       cleanup_close int subsystemfd = -1;
       subsystem = strchr (from, ':') + 1;
       subpath = strchr (subsystem, ':') + 1;
@@ -668,6 +677,9 @@ do_mount_cgroup_v1 (libcrun_container_t *container,
       if (subsystem[0] == '\0')
         continue;
 
+      /* subsystem_fqn includes name= for named hierarchies.  */
+      subsystem_fqn = subsystem;
+
       it = strstr (subsystem, "name=");
       if (it)
         subsystem += 5;
@@ -689,16 +701,35 @@ do_mount_cgroup_v1 (libcrun_container_t *container,
       if (UNLIKELY (ret < 0))
         return crun_make_error (err, errno, "open `%s`", subsystem_path);
 
-      ret = do_mount (container, source_path, subsystemfd, subsystem_path, NULL, MS_BIND | mountflags, NULL, 0, err);
-      if (UNLIKELY (ret < 0))
+      if (has_cgroupns)
         {
-          if (crun_error_get_errno (err) == ENOENT || crun_error_get_errno (err) == ENODEV)
+          ret = do_mount (container, source_path, subsystemfd, subsystem_path, "cgroup", mountflags, subsystem_fqn, 0, err);
+          if (UNLIKELY (ret < 0))
             {
-              /* We are trying to mount a subsystem that is not present.  */
+              if (crun_error_get_errno (err) == ENOENT || crun_error_get_errno (err) == ENODEV)
+                {
+                  /* We are trying to mount a subsystem that is not present.  */
+                  crun_error_release (err);
+                  continue;
+                }
+              return ret;
+            }
+        }
+      else
+        {
+          ret = do_mount (container, source_path, subsystemfd, subsystem_path, NULL, MS_BIND | mountflags, NULL, 0, err);
+          if (UNLIKELY (ret < 0))
+            {
+              if (crun_error_get_errno (err) != ENOENT)
+                return ret;
+
               crun_error_release (err);
-              continue;
+
+              /* We might already be in a container.  Mount the source subsystem.  */
+              ret = do_mount (container, source_subsystem, subsystemfd, subsystem_path, NULL, MS_BIND | mountflags, NULL, 0, err);
+              if (UNLIKELY (ret < 0))
+                return ret;
             }
-          return ret;
         }
     }
 
@@ -1004,7 +1035,7 @@ do_masked_or_readonly_path (libcrun_container_t *container,
       char source_buffer[64];
       sprintf (source_buffer, "/proc/self/fd/%d", pathfd);
 
-      ret = do_mount (container, source_buffer, pathfd, path, NULL, MS_BIND | MS_PRIVATE | MS_RDONLY | MS_REC, NULL, 0, err);
+      ret = do_mount (container, source_buffer, pathfd, path, NULL, MS_BIND | MS_PRIVATE | MS_RDONLY | MS_REC, NULL, true, err);
       if (UNLIKELY (ret < 0))
         return ret;
     }
@@ -1015,9 +1046,9 @@ do_masked_or_readonly_path (libcrun_container_t *container,
         return ret;
 
       if ((mode & S_IFMT) == S_IFDIR)
-        ret = do_mount (container, "tmpfs", pathfd, path, "tmpfs", MS_RDONLY, "size=0k", 0, err);
+        ret = do_mount (container, "tmpfs", pathfd, path, "tmpfs", MS_RDONLY, "size=0k", false, err);
       else
-        ret = do_mount (container, "/dev/null", pathfd, path, NULL, MS_BIND | MS_UNBINDABLE | MS_REC, NULL, 0, err);
+        ret = do_mount (container, "/dev/null", pathfd, path, NULL, MS_BIND | MS_UNBINDABLE | MS_REC, NULL, false, err);
       if (UNLIKELY (ret < 0))
         return ret;
     }
@@ -1050,7 +1081,7 @@ static int
 do_pivot (libcrun_container_t *container, const char *rootfs, libcrun_error_t *err)
 {
   int ret;
-  cleanup_close int oldrootfd = open ("/", O_DIRECTORY | O_RDONLY);
+  cleanup_close int oldrootfd = open ("/", O_DIRECTORY | O_PATH);
   cleanup_close int newrootfd = open (rootfs, O_DIRECTORY | O_RDONLY);
 
   if (UNLIKELY (oldrootfd < 0))
@@ -1347,7 +1378,14 @@ get_notify_fd (libcrun_context_t *context, libcrun_container_t *container, int *
   char *host_path = NULL;
 
   if (container)
-    host_path = get_private_data (container)->host_notify_socket_path;
+    {
+      const char *parent_dir;
+
+      parent_dir = get_private_data (container)->host_notify_socket_path;
+
+      xasprintf (&host_notify_socket_path, "%s/notify", parent_dir);
+      host_path = host_notify_socket_path;
+    }
 
   *notify_socket_out = -1;
 
@@ -1761,6 +1799,40 @@ libcrun_do_pivot_root (libcrun_container_t *container, bool no_pivot, const char
   return 0;
 }
 
+/* If one of stdin, stdout, stderr are pointing to /dev/null on
+ * the outside of the container, this moves it to /dev/null inside
+ * of the container. This needs to run afer pivot/chroot-ing. */
+int
+libcrun_reopen_dev_null (libcrun_error_t * err)
+{
+  struct stat dev_null;
+  struct stat statbuf;
+  cleanup_close int fd;
+  int i;
+
+  /* Open /dev/null inside of the container. */
+  fd = open ("/dev/null", O_RDWR);
+  if (UNLIKELY (fd == -1))
+    return crun_make_error (err, errno, "failed open()ing /dev/null");
+
+  if (UNLIKELY (fstat (fd, &dev_null) == -1))
+      return crun_make_error (err, errno, "failed stat()ing /dev/null");
+
+  for (i = 0; i <= 2; i++)
+    {
+      if (UNLIKELY (fstat (i, &statbuf) == -1))
+	  return crun_make_error (err, errno, "failed stat()ing fd %d", i);
+      if (statbuf.st_rdev == dev_null.st_rdev)
+	{
+	  /* This FD is pointing to /dev/null. Point it to /dev/null inside
+	   * of the container. */
+	  if (UNLIKELY (dup2 (fd, i) == -1))
+	    return crun_make_error (err, errno, "failed dup2()ing %d", i);
+	}
+    }
+  return 0;
+}
+
 static int
 uidgidmap_helper (char *helper, pid_t pid, char *map_file, libcrun_error_t *err)
 {
@@ -2243,9 +2315,8 @@ libcrun_set_oom (libcrun_container_t *container, libcrun_error_t *err)
 }
 
 int
-libcrun_set_sysctl (libcrun_container_t *container, libcrun_error_t *err)
+libcrun_set_sysctl_from_schema (runtime_spec_schema_config_schema *def, libcrun_error_t *err)
 {
-  runtime_spec_schema_config_schema *def = container->container_def;
   size_t i;
   cleanup_close int dirfd = -1;
 
@@ -2277,6 +2348,12 @@ libcrun_set_sysctl (libcrun_container_t *container, libcrun_error_t *err)
   return 0;
 }
 
+int
+libcrun_set_sysctl (libcrun_container_t *container, libcrun_error_t *err)
+{
+  return libcrun_set_sysctl_from_schema (container->container_def, err);
+}
+
 static int
 open_terminal (libcrun_container_t *container, char **slave, libcrun_error_t *err)
 {
@@ -2379,6 +2456,7 @@ libcrun_run_linux_container (libcrun_container_t *container,
   int namespaces_to_join_index[MAX_NAMESPACES];
   size_t n_namespaces_to_join = 0;
   int userns_join_index = -1;
+  bool must_fork = false;
 
   for (i = 0; i < def->linux->namespaces_len; i++)
     {
@@ -2455,6 +2533,13 @@ libcrun_run_linux_container (libcrun_container_t *container,
   if (UNLIKELY (ret < 0))
       goto out;
 
+  if (flags & CLONE_NEWPID)
+    must_fork = true;
+#ifdef CLONE_NEWTIME
+  if (flags & CLONE_NEWTIME)
+    must_fork = true;
+#endif
+
   /* If we create a new user namespace, create it as part of the clone.  */
   pid = syscall_clone ((flags & ((userns_join_index >= 0) ? 0 : CLONE_NEWUSER)) | (detach ? 0 : SIGCHLD), NULL);
   if (UNLIKELY (pid < 0))
@@ -2480,7 +2565,7 @@ libcrun_run_linux_container (libcrun_container_t *container,
             return crun_make_error (err, errno, "write to sync socket");
         }
 
-      if (flags & CLONE_NEWPID)
+      if (must_fork)
         {
           ret = TEMP_FAILURE_RETRY (read (sync_socket_host, &grandchild, sizeof (grandchild)));
           if (UNLIKELY (ret < 0))
@@ -2587,9 +2672,34 @@ libcrun_run_linux_container (libcrun_container_t *container,
         goto out;
     }
 
-  if (flags & CLONE_NEWPID)
+#ifdef CLONE_NEWTIME
+  if (flags & CLONE_NEWTIME)
     {
-      /* A PID namespace is joined when a new process is created.  */
+      cleanup_close int fd = open ("/proc/self/timens_offsets", O_WRONLY | O_CLOEXEC);
+      if (container->container_def->annotations)
+        {
+          size_t i;
+
+          for (i = 0; i < container->container_def->annotations->len; i++)
+            {
+              if (strcmp (container->container_def->annotations->keys[i], "run.oci.timens_offset") == 0)
+                {
+                  const char *v;
+
+                  v = container->container_def->annotations->values[i];
+
+                  ret = write (fd, v, strlen (v));
+                  if (UNLIKELY (ret < 0))
+                    goto out;
+                }
+            }
+        }
+    }
+#endif
+
+  if (must_fork)
+    {
+      /* A PID and a time namespace is joined when a new process is created.  */
       pid_container = fork ();
       if (UNLIKELY (ret < 0))
         {
@@ -2694,28 +2804,6 @@ join_process_parent_helper (pid_t child_pid,
   return pid;
 }
 
-static int
-inherit_env (pid_t pid_to_join, libcrun_error_t *err)
-{
-  int ret = 0;
-  size_t len;
-  char *str;
-  cleanup_free char *path;
-  /* Not a memory leak here.  The data used by putenv must not be freed.  */
-  char *content = NULL;
-
-  xasprintf (&path, "/proc/%d/environ", pid_to_join);
-
-  ret = read_all_file (path, &content, &len, err);
-  if (UNLIKELY (ret < 0))
-    return ret;
-
-  for (str = content; str < content + len; str += strlen (str) + 1)
-    if (putenv (str) < 0)
-        return crun_make_error (err, errno, "putenv `%s`", str);
-  return ret;
-}
-
 int
 libcrun_join_process (libcrun_container_t *container, pid_t pid_to_join, libcrun_container_status_t *status, int detach, int *terminal_fd, libcrun_error_t *err)
 {
@@ -2761,17 +2849,6 @@ libcrun_join_process (libcrun_container_t *container, pid_t pid_to_join, libcrun
   close_and_reset (&sync_socket_fd[0]);
   sync_fd = sync_socket_fd[1];
 
-  ret = clearenv ();
-  if (UNLIKELY (ret < 0))
-    {
-      crun_make_error (err, 0, "clearenv");
-      goto exit;
-    }
-
-  ret = inherit_env (pid_to_join, err);
-  if (UNLIKELY (ret < 0))
-    goto exit;
-
   if (def->linux->namespaces_len >= 10)
     {
       crun_make_error (err, 0, "invalid configuration");
@@ -2831,7 +2908,7 @@ libcrun_join_process (libcrun_container_t *container, pid_t pid_to_join, libcrun
   for (i = 0; all_namespaces[i]; i++)
     close_and_reset (&fds[i]);
 
-  if (detach && setsid () < 0)
+  if (setsid () < 0)
     {
       crun_make_error (err, errno, "setsid");
       goto exit;
@@ -3034,3 +3111,23 @@ __attribute__((constructor)) static void libcrun_rexec(void)
       _exit (EXIT_FAILURE);
     }
 }
+
+int
+libcrun_container_checkpoint_linux (libcrun_container_status_t *status,
+                                    libcrun_container_t *container,
+                                    libcrun_checkpoint_restore_t *cr_options,
+                                    libcrun_error_t *err)
+{
+  return libcrun_container_checkpoint_linux_criu (status, container,
+                                                  cr_options, err);
+}
+
+int
+libcrun_container_restore_linux (libcrun_container_status_t *status,
+                                 libcrun_container_t *container,
+                                 libcrun_checkpoint_restore_t *cr_options,
+                                 libcrun_error_t *err)
+{
+  return libcrun_container_restore_linux_criu (status, container,
+                                               cr_options, err);
+}
