@@ -4,7 +4,7 @@
  * Copyright (C) 2017, 2018, 2019 Giuseppe Scrivano <giuseppe@scrivano.org>
  * crun is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
- * the Free Software Foundation; either version 3 of the License, or
+ * the Free Software Foundation; either version 2.1 of the License, or
  * (at your option) any later version.
  *
  * crun is distributed in the hope that it will be useful,
@@ -441,15 +441,18 @@ crun_safe_ensure_at (bool dir, int dirfd, const char *dirpath, size_t dirpath_le
         {
           ret = openat (cwd, cur, O_CLOEXEC | O_CREAT | O_WRONLY, 0700);
           if (UNLIKELY (ret < 0))
-            return crun_make_error (err, errno, "create `%s`", path);
+            {
+              int saved_errno = errno;
+
+              /* If the previous openat fails, attempt to open the file in O_PATH mode.  */
+              ret = openat (cwd, cur, O_CLOEXEC | O_PATH);
+              if (ret < 0)
+                return crun_make_error (err, saved_errno, "create `%s`", path);
+            }
 
           close_and_replace (&wd_cleanup, ret);
 
-          ret = check_fd_under_path (dirpath, dirpath_len, ret, path, err);
-          if (UNLIKELY (ret < 0))
-            return ret;
-
-          return 0;
+          return check_fd_under_path (dirpath, dirpath_len, ret, path, err);
         }
 
       if (ret < 0)
